/*********************************************************************
**
** chkperm.c
** (c) 2003 Oliver Schroeder <progs@o-schroeder.de>
** www.o-schroeder.de
**
*********************************************************************/

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#include <dirent.h>
#include <libintl.h>
#include <locale.h>
#include "clist.h"

#define _(String) gettext (String)
#define MAX_INPUT_BUFFER    2048

/*********************************************************************
**
**      define type bool
**
*********************************************************************/
typedef enum
{
	FALSE   = 0,
	TRUE    = 1
} bool;

/*********************************************************************
**
**      constants (instead of #define's)
**
*********************************************************************/
typedef enum
{
	UnSpecified     = -1,
	NoRecursive     = 0,
	RecursiveFiles,
	RecursiveDirs,
	RecursiveAll,
	AllFiles,
	AllDirs,
	All
} t_ResursiveType;

/*********************************************************************
**
**      parsed from out config file
**
*********************************************************************/
typedef struct
{
	t_ResursiveType recursive_type; /* 0=no recursion, 1=dirs, 2=files */
	uid_t   owner;
	gid_t   group;
	char*   owner_name;
	char*   group_name;
	mode_t  permissions;
} my_permissions;

/*********************************************************************
**
**      global variables
**
*********************************************************************/
t_List* stack;
bool show_help          = FALSE;
bool set_permissions    = FALSE;
bool remove_dead_links  = FALSE;
bool be_verbose         = FALSE;
bool read_permissions   = FALSE;
bool follow_symlinks    = FALSE;
static char config_file[MAX_INPUT_BUFFER];

/*********************************************************************
**
**      prototypes
**
*********************************************************************/
void    parse_params ( int argc, char* argv [] );
void    display_help ();
void    process_line ();
void    process_config_file ();
int     chk_perms ( char* path, my_permissions* perms );
void    print_perms ( char* path );
int     recursive_dir ( char* path, my_permissions* perms );
int     check_path ( char* path );
int     str_cpy ( char**  ppsz_Dest, char*   psz_Src );
mode_t  char_to_mode ( char* modes, int* error, int base );

/*********************************************************************
**
**      main
**
*********************************************************************/
int
main ( int argc, char* argv [] )
{
	setlocale ( LC_ALL, "" );
	bindtextdomain ( PACKAGE, LOCALEDIR );
	textdomain ( PACKAGE );
	parse_params ( argc, argv );
	process_config_file ();
	return ( 0 );
} /* main() */

/*********************************************************************
**
**      parse command line arguments
**
*********************************************************************/
void
parse_params ( int argc, char* argv [] )
{
	int i;
	strncpy ( config_file, SYSCONF_LOC, sizeof ( SYSCONF_LOC ) -1 );
	strcat  ( config_file, "/permissions" );
	for ( i = 1; i < argc; i++ )
	{
		if ( ( ! strcmp ( argv[i], "-h" ) )
                ||  ( ! strcmp ( argv[i], "--help" ) ) )
		{
			show_help = TRUE;
		}
		else if ( ( ! strcmp ( argv[i], "-s" ) )
                ||  ( ! strcmp ( argv[i], "--set" ) ) )
		{
			set_permissions = TRUE;
		}
		else if ( ( ! strcmp ( argv[i], "-d" ) )
                ||  ( ! strcmp ( argv[i], "--dead" ) ) )
		{
			remove_dead_links = TRUE;
		}
		else if ( ( ! strcmp ( argv[i], "-r" ) )
                ||  ( ! strcmp ( argv[i], "--read" ) ) )
		{
			read_permissions = TRUE;
		}
		else if ( ( ! strcmp ( argv[i], "-v" ) )
                ||  ( ! strcmp ( argv[i], "--verbose" ) ) )
		{
			be_verbose = TRUE;
		}
		else if ( ( ! strcmp ( argv[i], "-f" ) )
                ||  ( ! strcmp ( argv[i], "--follow" ) ) )
		{
			follow_symlinks = TRUE;
		}
		else if ( argv [i][0] == '-' )
		{
			printf ( _("unknown option '%s'!\n"), argv [i] );
			show_help = TRUE;
		}
		else    /* must be the configfile name */
		{
			strncpy ( config_file, argv[i], sizeof ( config_file ) -1 );
		}
	} /* for */
	if ( show_help == TRUE )
	{
		display_help ();
	}
} /* parse_params() */

/*********************************************************************
**
**      display help screen and exit
**
*********************************************************************/
void
display_help ()
{
	printf ( "\nchkperm %s / (c) 2002-2008 Oliver Schroeder "
	         "<chkperm@o-schroeder.de>\n", VERSION );
	printf ( _("\nSyntax: chkperm [options] permission_file\n") );
	printf ( "\n" );
	printf ( _("permission_file    = file with owner/mode descriptions\n"));
	printf ( "\n" );
	printf ( _("options:\n") );
	printf ( _("-h|--help           = show this help\n" ) );
	printf ( _("-s|--set            = check and set file permissions\n") );
	printf ( _("-d|--dead           = delete dead symbolic links\n") );
	printf ( _("-r|--read           = read and print file permissions\n") );
	printf ( _("-f|--follow         = follow symlinks\n") );
	printf ( _("-v|--verbose        = be even more verbose\n") );
	printf ( "\n" );
	exit ( 0 );
} /* display_help */

/*********************************************************************
**
**      parse one line of the config file and call actions
**      accordingly.
**
*********************************************************************/
void
process_line ( int line, char* input_buffer )
{
	int     i;
	char*   token;
	char*   tmp_owner;
	char*   owner;
	char*   group;
	char*   path;
	char*   tmp_perm;
	int     field_no;
	my_permissions  perms;
	t_ResursiveType old_recursive_type = UnSpecified;
	struct passwd*  user_entry;
	struct group*   group_entry;
	field_no = 1;
	bzero ( &perms, sizeof ( perms ) );
	token = strtok ( input_buffer, " \t\n" );
	while ( token != NULL )
	{
		if ( field_no == 1 )    /* path or recursive_type */
		{
			old_recursive_type = perms.recursive_type;
			if ( !strcmp ( token, "recursivedirs" ) )
			{
				perms.recursive_type = RecursiveDirs;
				token = strtok ( NULL, " \t\n" );
			}
			else if ( !strcmp ( token, "recursivefiles" ) )
			{
				perms.recursive_type = RecursiveFiles;
				token = strtok ( NULL, " \t\n" );
			}
			else if ( !strcmp ( token, "recursiveall" ) )
			{
				perms.recursive_type = RecursiveAll;
				token = strtok ( NULL, " \t\n" );
			}
			else if ( !strcmp ( token, "allfiles" ) )
			{
				perms.recursive_type = AllFiles;
				token = strtok ( NULL, " \t\n" );
			}
			else if ( !strcmp ( token, "alldirs" ) )
			{
				perms.recursive_type = AllDirs;
				token = strtok ( NULL, " \t\n" );
			}
			else if ( !strcmp ( token, "all" ) )
			{
				perms.recursive_type = All;
				token = strtok ( NULL, " \t\n" );
			}
			else
			{
				perms.recursive_type = 0;
			}
			if ( token == NULL )
			{
				fprintf ( stderr,
					_("error: no path statement on line "
					"%u\n"), line );
				return;
			}
			path = token;
			if ( strchr ( path, '/' ) == 0 )
			{
				fprintf ( stderr,
					_("'%s' is an illegal statement on "
					"line %u\n"), path, line );
				return;
			}
		} /* if */
		else if ( field_no == 2 ) /* owner.group */
		{
			tmp_owner = token;
		}
		else if ( field_no == 3 ) /* permissions */
		{
			tmp_perm = token;
		}
		token = strtok ( NULL, " \t\n" );
		field_no++;
	} /* while */
	if ( strcmp ( tmp_perm, "-" ) == 0 )
	{
		if ( be_verbose )
		{
			printf ( _("permissions=0, leaving them untouched\n") );
		}
		perms.permissions = 0;
	}
	else
	{
		perms.permissions = char_to_mode ( tmp_perm, & i, 8 );
		if ( i != 0 )
		{
			fprintf ( stderr,
				_("illegal file modes on line %u\n"), line );
			return;
		}
	}
	owner = strtok ( tmp_owner, ":" );
	group = strtok ( NULL, "" );
	if ( strcmp ( owner, "-" ) != 0 )
	{
		user_entry = getpwnam ( owner );
		if ( user_entry == NULL )
		{
			fprintf ( stderr,
			          _("invalid user '%s' on line %u\n"),
				  owner, line );
			return;
		}
		perms.owner = user_entry->pw_uid;
		perms.owner_name = user_entry->pw_name;
	}
	else
	{
		if ( be_verbose )
		{
			printf ( _("user = 0, don't changing user ownership\n") );
		}
		perms.owner      = -1;
		perms.owner_name = 0;
		user_entry       = 0;
	}
	if ( strcmp ( group, "-" ) != 0 )
	{
		group_entry = getgrnam ( group );
		if ( group_entry == NULL )
		{
			fprintf ( stderr,
				_("invalid group '%s' on line %u\n"),
				group, line );
			return;
		}
		perms.group      = group_entry->gr_gid;
		perms.group_name = group_entry->gr_name;
	}
	else
	{
		if ( be_verbose )
		{
			printf ( _("group = 0, don't changing group ownership\n") );
		}
		perms.group      = -1;
		group_entry      = 0;
		perms.group_name = 0;
	}
	if ( perms.recursive_type > 0 )
	{
		if ( old_recursive_type != perms.recursive_type )
		{
			ListDestroy ( stack, TRUE );
			stack = ListCreate ();
		}
		if ( check_path ( path ) != 0 )
		{
			if ( be_verbose )
			{
				printf ( _("'%s' already processed, skipping\n"), path );
			}
			return;
		}
		i = recursive_dir ( path, & perms );
	}
	else
	{
		i = chk_perms ( path, & perms );
	}
	return;
} /* process_line() */

/*********************************************************************
**
**      process the config (permission description) file
**
*********************************************************************/
void
process_config_file ()
{
	FILE*   fp;
	char    input_buffer [MAX_INPUT_BUFFER];
	int     line;
	if ( be_verbose )
	{
		printf ( _("## reading '%s' ...\n"), config_file );
	}
	fp = fopen ( config_file, "r" );
	if ( fp == NULL )
	{
		fprintf ( stderr, _("Could not open '%s' for reading!\n"),
		          config_file );
		exit ( 1 );
	}
	line = 0;
	while ( fgets ( input_buffer, MAX_INPUT_BUFFER - 1, fp ) )
	{
		line++;
		if ( be_verbose )
		{
			printf ( "\n# %u: %s\n", line, input_buffer );
		}
		/* skip comments and blank lines */
		if ( ( input_buffer [0] == '#' )
		||  ( input_buffer [0] == '\0' )
		||  ( input_buffer [0] == '\n' ) )
		{
			continue;
		}
		process_line ( line, input_buffer );
	} /* while */
	ListDestroy ( stack, TRUE );
} /* process_config_file() */

/*********************************************************************
**
**      check if path is a broken link
**
*********************************************************************/
int
check_link ( char* path, struct stat* stats )
{
	errno = 0;
	lstat ( path, stats );
	if ( S_ISLNK ( stats->st_mode ) )
	{
		/* is a (broken) symbolic link */
		fprintf ( stderr, _("'%s' is a broken symbolic link... "),
			path );
		if ( errno == ELOOP )
		{
			fprintf ( stderr, _("'%s' too many symbolic links... "),
				path );
		}
		if ( remove_dead_links == TRUE )
		{
			fprintf ( stderr, _("deleteing") );
			unlink ( path );
		}
		fprintf ( stderr, "\n" );
		return ( 0 );
	}
	return ( 1 );
} /* check_link() */

/*********************************************************************
**
**      check if path was alread processed
**
*********************************************************************/
int
check_path ( char* path )
{
	char* current;
	if ( ListEmpty ( stack ) )
	{
		str_cpy ( & current, path );
		ListPushBack ( stack, current );
		return ( 0 );
	}
	ListSetStart ( stack );
	while ( ( current = ( char* ) ListNext ( stack ) ) != NULL )
	{
		if ( strcmp ( path, current ) == 0 )
		{
			return ( 1 );
		}
	}
	str_cpy ( & current, path );
	ListPushBack ( stack, current );
	return ( 0 );
} /* check_path() */

/*********************************************************************
**
**      iterate through dir and set permissions
**
*********************************************************************/
int
recursive_dir ( char* path, my_permissions* perms )
{
	int             error;
	char            complete_path [MAX_INPUT_BUFFER];
	DIR*            dirp;
	struct stat     stats;
	struct stat     link_stats;
	struct dirent*  dir_entry;
	dirp = opendir ( path );
	if ( dirp == NULL )
	{
		fprintf ( stderr, _("could not read directory '%s'!\n"), path );
		return ( -1 );
	}
	if ( be_verbose )
	{
		printf ( _("# processing directory %s:\n"), path );
	}
	if ( ( perms->recursive_type == RecursiveDirs )
	||  ( perms->recursive_type == RecursiveAll )
	||  ( perms->recursive_type == AllDirs ) )
	{
		/* check directory itself */
		chk_perms ( path, perms );
	}
	while ( ( dir_entry = readdir ( dirp ) ) != NULL )
	{
		if ( ( ! strcmp ( dir_entry->d_name, "." ) )
		||  ( ! strcmp ( dir_entry->d_name, ".." ) ) )
		{
			continue;
		}
		/*
		** dir_entry->d_type is not always implemented
		** so stat() it *uff*
		*/
		strcpy ( complete_path, path );
		if ( path[strlen ( path ) - 1] != '/' )
		{
			strcat ( complete_path, "/" );
		}
		strcat ( complete_path, dir_entry->d_name );
		errno = 0;
		error = stat ( complete_path, & stats );
		if ( error != 0 )
		{
			if ( !check_link ( complete_path, & stats ) )
			{
				continue;
			}
			fprintf ( stderr, "%s -> %s\n",
			          complete_path, strerror ( errno ) );
			continue;
		}
		if ( S_ISDIR ( stats.st_mode )
		                && ( check_path ( complete_path ) != 0 ) )
		{
			if ( be_verbose )
			{
				printf ( _("%s alread processed, skipping\n"),
				         complete_path );
			}
			continue;
		}
		lstat ( complete_path, &link_stats );
		switch ( perms->recursive_type )
		{
		case RecursiveDirs:
			if ( S_ISDIR ( stats.st_mode ) )
			{
				if ( ( S_ISLNK ( link_stats.st_mode ) )
				                && ( ! follow_symlinks ) )
				{
					if ( be_verbose )
					{
						printf ( _("# not following symbolic link '%s'\n"),
						         complete_path );
					}
				}
				else
				{
					recursive_dir ( complete_path, perms );
				}
			}
			break;
		case RecursiveFiles:
		case RecursiveAll:
			if ( S_ISDIR ( stats.st_mode ) )
			{
				if ( ( S_ISLNK ( link_stats.st_mode ) )
				                && ( ! follow_symlinks ) )
				{
					if ( be_verbose )
					{
						printf ( _("# not following symbolic link '%s'\n"),
						         complete_path );
					}
				}
				else
				{
					recursive_dir ( complete_path, perms );
				}
			}
			else
			{
				chk_perms ( complete_path, perms );
			}
			break;
		case AllDirs:
			if ( S_ISDIR ( stats.st_mode ) )
			{
				chk_perms ( complete_path, perms );
			}
			break;
		case AllFiles:
			if ( ! S_ISDIR ( stats.st_mode ) )
			{
				chk_perms ( complete_path, perms );
			}
			break;
		case All:
			chk_perms ( complete_path, perms );
			break;
		default:
			/* keep compilers happy */
			break;
		} /* switch */
	}
	closedir ( dirp );
	if ( be_verbose )
	{
		printf ( _("# leaving recursive directory %s:\n"), path );
	}
	return ( 0 );
} /* recursive_dir() */

/*********************************************************************
**
**      read permissions and owner and print it in a way
**      chkperm can read as input
**
*********************************************************************/
void
print_perms ( char* path )
{
	struct stat     stats;
	extern int      errno;
	int             error;
	mode_t          file_permissions;
	struct passwd*  user_entry;
	struct group*   group_entry;
	errno = 0;
	error = stat ( path, & stats );
	if ( be_verbose )
	{
		printf ( _("# reading file permissions for '%s'\n"), path );
	}
	if ( error != 0 )
	{
		if ( !check_link ( path, & stats ) )
		{
			return;
		}
		fprintf ( stderr, "%s -> %s\n", path, strerror ( errno ) );
		return;
	}
	user_entry = getpwuid ( stats.st_uid );
	group_entry  = getgrgid ( stats.st_gid );
	file_permissions = ( stats.st_mode & 0007777 );
	printf ( "%-55s %s:%s\t%o\n", path, user_entry->pw_name,
	         group_entry->gr_name, file_permissions );
} /* print_perms () */

/*********************************************************************
**
**      check permissions and modify accordingly
**
*********************************************************************/
int
chk_perms ( char* path, my_permissions* perms )
{
	struct stat     stats;
	extern int      errno;
	int             error;
	bool            do_change;
	mode_t          file_permissions;
	struct passwd*  user_entry;
	struct group*   group_entry;

	if ( read_permissions == TRUE )
	{
		print_perms ( path );
		return ( 0 );
	}
	if ( be_verbose )
	{
		printf ( _("# checking '%s'\n"), path );
	}
	errno = 0;
	error = stat ( path, & stats );
	if ( error != 0 )
	{
		if ( !check_link ( path, & stats ) )
		{
			return ( 0 );
		}
		fprintf ( stderr, "%s -> %s\n", path, strerror ( errno ) );
		return ( -1 );
	}
	if ( ! S_ISDIR ( stats.st_mode ) )
	{
		if ( stats.st_nlink > 1 )
		{
			fprintf ( stderr, _("'%s' has to much hard links (%i). Ignored.\n"),
				path, ( int ) stats.st_nlink );
			return 0;
		}
	}
	do_change = FALSE;
	if ( perms->owner != -1 )
	{
		if ( perms->owner != stats.st_uid )
		{
			errno = 0;
			user_entry = getpwuid ( stats.st_uid );
			if ( user_entry == 0 )
			{
				printf ( _("'%s' owner is invalid, has to be %s\n"), path,
				         perms->owner_name );
			}
			else
			{
				printf ( _("'%s' owner is %s, has to be %s\n"),
					path,
					user_entry->pw_name, perms->owner_name);
			}
			do_change = TRUE;
		}
	}
	if ( perms->group != -1 )
	{
		if ( perms->group != stats.st_gid )
		{
			group_entry  = getgrgid ( stats.st_gid );
			if ( group_entry == 0 )
			{
				printf ( _("'%s' group is invalid, has to be %s\n"), path,
					perms->group_name );
			}
			else
			{
				printf ( _("'%s' group is %s, has to be %s\n"), path,
					group_entry->gr_name, perms->group_name );
			}
			do_change = TRUE;
		}
	}
	errno = 0;
	if ( ( do_change == TRUE ) && ( set_permissions == TRUE ) )
	{
		printf ( _("changing owner\n") );
		errno = 0;
		error = chown ( path, perms->owner, perms->group );
		if ( error != 0 )
		{
			perror ( strerror ( errno ) );
		}
	}
	if ( error )
	{
		fprintf ( stderr, "%s -> %s\n", path, strerror ( errno ) );
		return ( -1 );
	}
	if ( perms->permissions == 0 )
	{
		return ( 0 );
	}
	/* delete file type bitfield */
	file_permissions = ( stats.st_mode & 0007777 );
	if ( file_permissions != perms->permissions )
	{
		/* set original file type bitfield */
		file_permissions = ( stats.st_mode & ~0007777 );
		/* set normal file permissions */
		file_permissions |= perms->permissions;
		printf ( _("'%s' file permissions are %o, have to be %o\n"), path,
			( int ) ( stats.st_mode & 0007777 ),
			( int ) ( file_permissions & 0007777 ) );
		errno = 0;
		if ( set_permissions == TRUE )
		{
			printf ( _("changing file permission\n") );
			errno = 0;
			error = chmod ( path, file_permissions );
			if ( error != 0 )
			{
				perror ( strerror ( errno ) );
			}
		}
		if ( error )
		{
			fprintf ( stderr, "%s -> %s\n", path, strerror ( errno ) );
			return ( -1 );
		}
	}
	return ( 0 );
} /* chk_perms() */

/*********************************************************************
**
**      allocate memory for 'dest' and copy src to it
**
*********************************************************************/
int
str_cpy ( char**  dest, char*   src )
{
	int     len;
	char*   newstr;
	if ( src == 0 )
	{
		return ( 0 );
	}
	len   = strlen ( src );
	newstr = ( char* ) malloc ( len + 1 );
	if ( newstr == 0 )
	{
		return ( -1 );
	}
	memset ( newstr, 0, len + 1 );
	strncpy ( newstr, src, len );
	( *dest ) = newstr;
	return ( 0 );
} /* str_cpy() */

/*********************************************************************
**
**      convert a mode-string ("1775") to a mode_t
**
*********************************************************************/
mode_t
char_to_mode ( char* modes, int* error, int base )
{
	mode_t          result;
	int             index;
	int             length;
	int             current;
	result = 0;
	length = strlen ( modes );
	if ( length <= 0 )
	{
		( *error ) = -1;
		return ( result );
	}
	index = 0;
	( *error ) = 0;
	while ( index < length )
	{
		current = modes [index] - '0';
		if ( ( current < 0 ) || ( current > base - 1 ) )
		{
			fprintf ( stderr, _("character out of range: "
				"char: '%c'  integer: %u  base: %u\n"),
				modes [index], current, base );
			( *error ) = index + 1;
			return ( result );
		}
		result *= base;
		result += current;
		index++;
	} /* while */
	return ( result );
} /* char_to_mode() */
